/******************************************************************************
 *  Echo Server v1.2.0
 *  Copyright (C) 2002 John David Ratliff and Christopher Michael O'Donnell
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifndef _WIN32
#include <unistd.h>

#include <sys/socket.h>
#include <netinet/in.h>
#else
#include <windows.h>
#include <winsock.h>
#endif

// the default server port
#define	DEFAULT_PORT	64201

// startServer()
// 	returns the server socket, or -1 if we couldn't bind the socket
int startServer(int port) {
	int sockfd;
	struct sockaddr_in address;
#ifndef _WIN32
	int yes = 1;
#else
	char yes = 1;
	WORD wVersionRequested = MAKEWORD(2,2);
	WSADATA wsaData;
#endif

#ifdef _WIN32
	// initialize the Windows Winsockets library
	if (WSAStartup(wVersionRequested,&wsaData) != 0) {
		fprintf(stderr,"Unable to initialize the Winsock library!\n");
		return -1;
	}
#endif

	// open the server socket
	if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
		fprintf(stderr,"Unable to open socket!\n");
		return -1;
	}

	// setup bind stuffs
	address.sin_family = AF_INET;
	address.sin_addr.s_addr = INADDR_ANY;
	//address.sin_port = htons(port);
	address.sin_port = htons((unsigned short)port);

	// bind the socket
	if (bind(sockfd, (struct sockaddr *)&address, sizeof(struct sockaddr)) == -1) {
		fprintf(stderr,"Unable to bind socket!\n");
		return -1;
	}

	// allow socket rebinding of local ports
	if (setsockopt(sockfd,SOL_SOCKET,SO_REUSEADDR,&yes,sizeof(int)) == -1) {
		fprintf(stderr,"Unable to set reuseable socket binding property...\n");
	}

	// listen for connections
	if (listen(sockfd,10) != 0) {
		fprintf(stderr,"Unable to set socket for listen mode!\n");
		return -1;
	}

	// return the socket
	return sockfd;
}

// handleConnection()
// 	returns 1 for normal messages, and -1 if we received a shutdown message
short int handleConnection(int sockfd) {
	char buffer[1024];
	short int message = 1;

	// erase the buffer
	memset(buffer,0,1024);

	// read the client message
	recv(sockfd, buffer, 1023, 0);

	// print the client message and what we're doing
	printf("Client said \"%s\". Echoing...\n", buffer);

	// echo the client message
	send(sockfd, buffer, strlen(buffer), 0);

	// close the client socket
#ifndef _WIN32
	close(sockfd);
#else
	closesocket(sockfd);
#endif

	// check for shutdown message
	// we'll need a more secure shutdown method in the future...
	if (strcmp(buffer,"...shutdown now !@!") == 0) {
		printf("\nReceived shutdown message...\nStopping server...\n\n");
		message = -1;
	}

	// return -1 on shutdown, 1 otherwise
	return message;
}

// waitForConnections()
// 	waits in an infinite loop for sockect connections, then passes them to handleConnection
void waitForConnections(int sockfd) {
	struct sockaddr_in remoteAddress;
	int remoteAddressSize = sizeof(struct sockaddr_in);
	int remoteSocket;

	// listen for connections for...ever (or until we get a shutdown message)
	for (;;) {
		// accept the remote connection
		remoteSocket = accept(sockfd, (struct sockaddr *)&remoteAddress, &remoteAddressSize);

		// handle the remote client connection
		if (handleConnection(remoteSocket) == -1) {
			// if we got a shutdown message, break the loop
			break;
		}
	}
}

// main() - start here
// 	return EXIT_SUCCESS if we shutdown proper
int main(int argc, char **argv) {
	int sockfd = 0, port;

	if (argc > 1) {
		// user specified a port
		port = atoi(argv[1]);
	} else {
		// otherwise, use the default port
		port = DEFAULT_PORT;
	}

	if ((sockfd = startServer(port)) != -1) {
		// wait for incoming connections
		waitForConnections(sockfd);

		// close the server socket
#ifndef _WIN32
		close(sockfd);
#else
		closesocket(sockfd);

		// perform Winsock cleanup
		WSACleanup();
#endif
	}

	return EXIT_SUCCESS;
}
